/*
 *  $Id: PTK_driver_1bpp.c,v 1.3 2000/12/03 12:14:28 tgkamp Exp $
 *
 *  picoTK 1 bit per pixe1 driver, 2 colors (b&w)
 *
 *  Copyright (c) by Thomas Gallenkamp <tgkamp@users.sourceforge.net>
 *
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


#include <stdio.h>
#include "PTK.h"
#ifndef PICO_TOTAL_X
 #define PICO_TOTAL_X 640
#endif
#ifndef PICO_X
 #define PICO_X 640
#endif
#ifndef PICO_Y
 #define PICO_Y 480
#endif

#define LEFTMASK  "\x00\x80\xc0\xe0\xf0\xf8\xfc\xfe"
#define RIGHTMASK "\xff\x7f\x3f\x1f\x0f\x07\x03\x01"
#define R2MASK    "\x7f\x3f\x1f\x0f\x07\x03\x01\x00"
#define L2MASK    "\x80\xc0\xe0\xf0\xf8\xfc\xfe\xff"


int picoInfoSizeX(void)      { return PICO_X; }

int picoInfoSizeY(void)      { return PICO_Y; }

int picoInfoColorDepth(void) { return 1; }

inline void picoDrawPointRaw(int x, int y,int color) {
  unsigned char *p;
  p=fbuf+x/8+y*PICO_TOTAL_X/8;
  *p=color ?
     *p | (0x80>>(x&7)):
     *p & (0xff-(0x80>>(x&7)));
}

int picoDrawPixmapRaw (int x, int y, struct picoPixmap *pixmap, 
                       unsigned char fgc,unsigned char  bgc) {
    unsigned char *dst;
    unsigned char *src;
    int height;
    int width;
    int width_bytes;
    int xi, yi;
    int xm8,xm8m7;

    /*
     * We can only do depth 1 pixmaps
     */
    if (pixmap->type!=PICO_PIXMAP_1BPP) return 0;

    xm8=x&7;       /* x modulo 8 */
    xm8m7=8-(x&7); /* x modulo 8 minus 7 */


    dst=fbuf+y*PICO_TOTAL_X/8+x/8;
    height=pixmap->height_msb*256+pixmap->height_lsb;
    width =pixmap->width_msb *256+pixmap->width_lsb;
    width_bytes=(width+7)/8;
    src=pixmap->data;

    if (xm8==0) {
     /* speedcode */
     for (yi=0; yi<height; yi++) {
      for (xi=0; xi<width_bytes; xi++) {
        dst[xi]=*(src++);
      }
      dst+=PICO_TOTAL_X/8;
     }
    } else {
      /* generic code */
      unsigned char data,data_old;
      for (yi=0; yi<height; yi++) {
        data_old=dst[0]&LEFTMASK[xm8];
        for (xi=0; xi<width_bytes; xi++) {
         data=*(src++);
         dst[xi]=(data>>xm8)+data_old;
         data_old=data<<xm8m7;       
        }
        dst[xi]=data_old+(dst[xi]&RIGHTMASK[xm8]);
        dst+=PICO_TOTAL_X/8;
      }
    }
}

void picoFillRectRaw(int x, int y, int w, int h, int color) {
  unsigned char *dst;
  int width_bytes;
  int xmf, xmt, xf, xt;
  int xi;

    
  xf=x/8; xt=(x+w-1)/8;
  xmf=x&7; xmt=(x+w-1)&7;
  for (h+=y; y<h; y++) {
   unsigned char mask;
   dst=fbuf+y*PICO_TOTAL_X/8;
   if (xf==xt) {
     /* special case: both pixels in same byte */
     
    mask=LEFTMASK[xmf]|R2MASK[xmt];
    dst[xf]=(dst[xf]&mask)
            |((mask^0xff)&(color? 0xff:0));
   } else {
     mask=LEFTMASK[xmf];
     dst[xf]=(dst[xf]&mask)
            |((mask^0xff)&(color? 0xff:0));
     for (xi=xf+1; xi<xt; xi++) dst[xi]=color? 0xff:0;
     mask=R2MASK[xmt];
     dst[xt]=(dst[xt]&mask)
            |((mask^0xff)&(color? 0xff:0));
   }
  }
}

void picoReverseRectRaw(int x, int y, int w, int h) {
  unsigned char *dst;
  int width_bytes;
  int xmf, xmt, xf, xt;
  int xi;

    
  xf=x/8; xt=(x+w-1)/8;
  xmf=x&7; xmt=(x+w-1)&7;
  for (h+=y; y<h; y++) {
   unsigned char mask;
   dst=fbuf+y*PICO_TOTAL_X/8;
   if (xf==xt) {
     /* special case: both pixels in same byte */    
    mask=LEFTMASK[xmf]|R2MASK[xmt];
    dst[xf]=dst[xf]^(mask^0xff);
   } else {
     dst[xf]=dst[xf]^(LEFTMASK[xmf]^0xff);
     for (xi=xf+1; xi<xt; xi++) dst[xi]=dst[xi]^0xff;
     dst[xt]=dst[xt]^(R2MASK[xmt]^0xff);
   }
  }
}

int picoDrawTextRaw(int x, int y, char *txt, struct picoFont *font,
                    int fgcolor, int bgcolor) {
  int ch;
  int cont;
  int x_orig;
  
  x_orig=x;
  if (font==NULL) return 0;
  cont=1;
  while (ch=(unsigned char)*txt) {
    int height;
    int width;
    int width_in_bytes;
    int yi,xi,xp;
    unsigned char *src;
    unsigned char *dst;
    unsigned char next_byte;
    unsigned char lrmask;
    unsigned char xm8;
    unsigned char src_data;
    height=font->height;
    width=font->per_char[ch].width;
    if (x+width>PICO_X) {
        cont=0;
        break;
    }
    width_in_bytes=(width+7)/8;
    src =font->data+font->per_char[ch].offset_msb*256
                        +font->per_char[ch].offset_lsb;
    dst=fbuf+y*PICO_TOTAL_X/8+x/8;
   if ((x&7)==0 && width==8) {
       /* speeded up code for 8 pixel font on byte boundary */
     if (fgcolor) 
       for (yi=0; yi<height; yi++) { *dst=*src;
                                     src++;
                                     dst+=PICO_TOTAL_X/8; }
     else 
       for (yi=0; yi<height; yi++) { *dst=*src^0xff;
                                      src++;
                                      dst+=PICO_TOTAL_X/8;
                                   }
   } else {
      /* generic code */
    xm8=x&7;
    for (yi=0; yi<height; yi++) {
#if 0
     next_byte=0;
     for (xi=0,xp=width;xp+xm8>0; xi++,xp-=8) {
       lrmask= ((xi==0) ? LEFTMASK[xm8]:0 )
              |((xp+xm8>=8) ? 0:RIGHTMASK[xp+xm8] );
       src_data=(fgcolor ? src[xi]:(src[xi]^(0xff)))
               &(xp>=8 ?0xff:(xp<=0 ? 0:LEFTMASK[xp]));  
       dst[xi]=(dst[xi]&lrmask)
              |((src_data>>xm8))
              |next_byte;
       next_byte=(src_data<<(8-xm8));
      } /* rof xi */
#else
   /* unrolled version of the above */
     xp=width;
     lrmask=LEFTMASK[xm8]
            |((xp+xm8>=8) ? 0:RIGHTMASK[xp+xm8] );
     src_data=fgcolor ? src[0]:(src[0]^(xp>=8 ?0xff:(LEFTMASK[xp])));
     dst[0]=(dst[0]&lrmask)|(src_data>>xm8);
     next_byte=(src_data<<(8-xm8));
     for (xi=1,xp-=8;xi<width_in_bytes; xi++,xp-=8) {
       lrmask=((xp+xm8>=8) ? 0:RIGHTMASK[xp+xm8] );
       src_data=fgcolor ? src[xi]:(src[xi]^(xp>=8 ?0xff:(LEFTMASK[xp])));
       dst[xi]=(dst[xi]&lrmask)
              |(src_data>>xm8)
              |next_byte;
       next_byte=(src_data<<(8-xm8));
     } /* rof xi */
     if (xp+xm8>0) dst[xi]=(dst[xi]&RIGHTMASK[xp+xm8])|next_byte;
   
#endif
     src+=width_in_bytes;
     dst+=PICO_TOTAL_X/8;
     } /* rof yi */
    } /* fi generic code */
    x+=width;
    txt++;
  } /* elihw ch */
 return x-x_orig;
}

int picoScrollRaw(int x, int y, int w, int h, int delta, int bgc) {
    unsigned char *src;
    unsigned char *dst;
    int xi,yi;

    int width_bytes;
    int xmf, xmt, xf, xt;

    xf=x/8; xt=(x+w-1)/8;
    xmf=x&7; xmt=(x+w-1)&7;
    
    if (delta>0) {
        dst=fbuf+y*PICO_TOTAL_X/8;
        src=dst+delta*PICO_TOTAL_X/8;
        for (yi=0; yi<h-delta; yi++) {
         if (xf==xt) {
          dst[xf]=(((LEFTMASK[xmf])|(R2MASK[xmt]))&dst[xf])
                 |(((RIGHTMASK[xmf])|(L2MASK[xmt]))&src[xf]);
         } else {
          dst[xf]=(LEFTMASK[xmf]&dst[xf])|(RIGHTMASK[xmf]&src[xf]);
          for (xi=xf+1; xi<xt; xi++) dst[xi]=src[xi];
          dst[xt]=(R2MASK[xmt]&dst[xt])|L2MASK[xmt]&src[xt];
         }
         dst+=PICO_TOTAL_X/8;
         src+=PICO_TOTAL_X/8;
        }
        for (; yi<h; yi++) {
            if (xf==xt) {
             dst[xf]=(((LEFTMASK[xmf])|(R2MASK[xmt]))&dst[xf]);
            } else {
             dst[xf]=LEFTMASK[xmf]&dst[xf];
             for (xi=xf+1; xi<xt; xi++) dst[xi]=0;
             dst[xt]=R2MASK[xmt]&dst[xt];
            }
         dst+=PICO_TOTAL_X/8;
        }
    } else {
        /*
         * not yet tested ...
         */
        dst=fbuf+(y+h-1)*PICO_TOTAL_X/8;
        src=dst-delta*PICO_TOTAL_X/8;
        for (yi=h-1; yi>=h-delta; yi--) {
            if (xf==xt) {
             dst[xf]=(((LEFTMASK[xmf])|(R2MASK[xmt]))&dst[xf])
                    |(((RIGHTMASK[xmf])|(L2MASK[xmt]))&src[xf]);
            } else {
             dst[xf]=(LEFTMASK[xmf]&dst[xf])|(RIGHTMASK[xmf]&src[xf]);
             for (xi=xf+1; xi<xt; xi++) dst[xi]=src[xi];
             dst[xt]=(R2MASK[xmt]&dst[xt])|L2MASK[xmt]&src[xt];
            }
            dst+=PICO_TOTAL_X/8;
            src+=PICO_TOTAL_X/8;
        }
        for (; yi<=0; yi--) {
            if (xf==xt) {
             dst[xf]=(((LEFTMASK[xmf])|(R2MASK[xmt]))&dst[xf]);
            } else {
             dst[xf]=LEFTMASK[xmf]&dst[xf];
             for (xi=xf+1; xi<xt; xi++) dst[xi]=0;
             dst[xt]=R2MASK[xmt]&dst[xt];
            }
         dst+=PICO_TOTAL_X/8;
        }
    }
}

void picoDriverInit(void) {
    /* Clear Screen */
    picoFillRectRaw(0,0,PICO_X,PICO_Y,0);
}
