/*
 *  $Id: PTK_driver_4bpp.c,v 1.2 2000/12/03 12:14:28 tgkamp Exp $
 *
 *  picoTK 4 bit per pixel driver, 16 colors
 *
 *  Copyright (c) by Thomas Gallenkamp <tgkamp@users.sourceforge.net>
 *
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


#include <stdio.h>
#include "PTK.h"


/* Change the below to adapt driver - do this in the Makefile rather
   than here. The defaultsettings here are for compatibility with VGA
   hardware (640x480)
 */

#ifndef PICO_TOTAL_X
 #define PICO_TOTAL_X 640
#endif
#ifndef PICO_X
 #define PICO_X 640
#endif
#ifndef PICO_Y
 #define PICO_Y 480
#endif

 
#define LEFTMASK  "\x00\xf0"
#define RIGHTMASK "\xff\x0f"
#define R2MASK    "\x0f\x00"
#define L2MASK    "\xf0\xff"

int picoInfoSizeX(void)      { return PICO_X; }

int picoInfoSizeY(void)      { return PICO_Y; }

int picoInfoColorDepth(void) { return 4; }


inline void picoDrawPointRaw(int x, int y,int color) {
  unsigned char *p;
  p=fbuf+x/2+y*PICO_TOTAL_X/2;
  *p=(x&1) ? ((*p & 0xf0) | color)
           : ((*p & 0x0f) | (color<<4));
}

int picoDrawPixmapRaw (int x, int y, struct picoPixmap *pixmap,
                       unsigned char fgc, unsigned char bgc) {
    unsigned char *dst;
    unsigned char *src;
    int height;
    int width;
    int width_bytes;
    int xi, yi;
    int xm8,xm8m7;

    /*
     * We can only do depth 1 pixmaps
     */
    if (pixmap->type!=PICO_PIXMAP_1BPP) return 0;
   
    dst=fbuf+y*PICO_TOTAL_X/2+x/2;
    height=pixmap->height_msb*256+pixmap->height_lsb;
    width =pixmap->width_msb *256+pixmap->width_lsb;
    width_bytes=(width+7)/8;
    src=pixmap->data;

    for (yi=0; yi<height; yi++) {
     unsigned char *p;
     unsigned char mem;
     p=dst;
     if (x&1) mem=*p & 0xf0;
     for (xi=0; xi<width; ) {
      unsigned char data;
      int i;
      data=*(src++);
      for (i=0; (i<8) && xi<width; xi++,i++) {
        unsigned char color;
        color=((data<<i) & 0x80) ? fgc:bgc;
        if ((xi+x)&1) {
          *p=mem|color;
          p++;
        } else {
          mem=(color<<4);
        }
      }/* i */
     } /* xi */
     if ((xi+x)&1) *p=(*p&0x0f)|mem;
     dst+=PICO_TOTAL_X/2;
    } 
}

void picoFillRectRaw(int x, int y, int w, int h, int color) {
  unsigned char *dst;
  int width_bytes;
  int xmf, xmt, xf, xt;
  int xi;

    
  xf=x/2; xt=(x+w-1)/2;
  xmf=x&2; xmt=(x+w-1)&2;
  for (h+=y; y<h; y++) {
   unsigned char mask;
   dst=fbuf+y*PICO_TOTAL_X/2;
   if (xf==xt) {
     /* special case: both pixels in same byte */
     
    mask=LEFTMASK[xmf]|R2MASK[xmt];
    dst[xf]=(dst[xf]&mask)
            |((mask^0xff)&(color*0x11));
   } else {
     mask=LEFTMASK[xmf];
     dst[xf]=(dst[xf]&mask)
            |((mask^0xff)&(color*0x11));
     for (xi=xf+1; xi<xt; xi++) dst[xi]=color*0x11;
     mask=R2MASK[xmt];
     dst[xt]=(dst[xt]&mask)
            |((mask^0xff)&(color*0x11));
   }
  }
}

void picoReverseRectRaw(int x, int y, int w, int h) {
  unsigned char *dst;
  int width_bytes;
  int xmf, xmt, xf, xt;
  int xi;

    
  xf=x/2; xt=(x+w-1)/2;
  xmf=x&1; xmt=(x+w-1)&1;
  for (h+=y; y<h; y++) {
   unsigned char mask;
   dst=fbuf+y*PICO_TOTAL_X/2;
   if (xf==xt) {
     /* special case: both pixels in same byte */    
    mask=LEFTMASK[xmf]|R2MASK[xmt];
    dst[xf]=dst[xf]^(mask^0xff);
   } else {
     dst[xf]=dst[xf]^(LEFTMASK[xmf]^0xff);
     for (xi=xf+1; xi<xt; xi++) dst[xi]=dst[xi]^0xff;
     dst[xt]=dst[xt]^(R2MASK[xmt]^0xff);
   }
  }
}

int picoDrawTextRaw(int x, int y, char *txt, struct picoFont *font,
                    int fgc, int bgc) {
  int ch;
  int cont;
  int x_orig;
  
  x_orig=x;
  if (font==NULL) return 0;
  cont=1;
  while (ch=(unsigned char)*txt) {
    int height;
    int width;
    int width_in_bytes;
    int yi,xi,xp;
    unsigned char *src;
    unsigned char *dst;
    unsigned char lrmask;
    unsigned char xm2;
    unsigned char src_data;
    height=font->height;
    width=font->per_char[ch].width;
    if (x+width>PICO_X) {
        cont=0;
        break;
    }
    width_in_bytes=(width+7)/8;
    src =font->data+font->per_char[ch].offset_msb*256
                        +font->per_char[ch].offset_lsb;
    dst=fbuf+y*PICO_TOTAL_X/2+x/2;
      /* generic code */
    xm2=x&1;
    for (yi=0; yi<height; yi++) {
     unsigned char *p;
     unsigned char mem;
     p=dst;
     if (x&1) mem=*p&0xf0;
     
     for (xi=0;xi<width; ) {
      unsigned char data;
      int i;
      data=*(src++);
      for (i=0; i<8 && xi<width; xi++, i++) {
         unsigned char color;
         color=((data<<i) & 0x80) ? fgc:bgc;
         if ((xi+x)&1) {
            *p=mem|color;
            p++;
         } else {
            mem=(color<<4);
         }
      } /* rof i */
     } /* rof xi */
     if ((xi+x)&1) *p=(*p&0x0f)|mem;
     dst+=PICO_TOTAL_X/2;
    } /* rof yi */
    x+=width;
    txt++;
  } /* elihw ch */
 return x-x_orig;
}


int picoScrollRaw(int x, int y, int w, int h, int delta, int bgc) {
    unsigned char *src;
    unsigned char *dst;
    int xi,yi;

    int width_bytes;
    int xmf, xmt, xf, xt;

    xf=x/2; xt=(x+w-1)/2;
    xmf=x&1; xmt=(x+w-1)&1;
    
    if (delta>0) {
        dst=fbuf+y*PICO_TOTAL_X/2;
        src=dst+delta*PICO_TOTAL_X/2;
        for (yi=0; yi<h-delta; yi++) {
         if (xf==xt) {
          dst[xf]=(((LEFTMASK[xmf])|(R2MASK[xmt]))&dst[xf])
                 |(((RIGHTMASK[xmf])|(L2MASK[xmt]))&src[xf]);
         } else {
          dst[xf]=(LEFTMASK[xmf]&dst[xf])|(RIGHTMASK[xmf]&src[xf]);
          for (xi=xf+1; xi<xt; xi++) dst[xi]=src[xi];
          dst[xt]=(R2MASK[xmt]&dst[xt])|L2MASK[xmt]&src[xt];
         }
         dst+=PICO_TOTAL_X/2;
         src+=PICO_TOTAL_X/2;
        }
        for (; yi<h; yi++) {
            if (xf==xt) {
             dst[xf]=(((LEFTMASK[xmf])|(R2MASK[xmt]))&dst[xf]);
            } else {
             unsigned char pattern;
             pattern=bgc*0x11;
             dst[xf]=(LEFTMASK[xmf]&dst[xf])|(RIGHTMASK[xmf]&pattern);
             for (xi=xf+1; xi<xt; xi++) dst[xi]=pattern;
             dst[xt]=(R2MASK[xmt]&dst[xt])|(L2MASK[xmt]&pattern);
            }
         dst+=PICO_TOTAL_X/2;
        }
    } else {
        /*
         * not yet tested ...
         */
        dst=fbuf+(y+h-1)*PICO_TOTAL_X/2;
        src=dst-delta*PICO_TOTAL_X/2;
        for (yi=h-1; yi>=h-delta; yi--) {
            if (xf==xt) {
             dst[xf]=(((LEFTMASK[xmf])|(R2MASK[xmt]))&dst[xf])
                    |(((RIGHTMASK[xmf])|(L2MASK[xmt]))&src[xf]);
            } else {
             dst[xf]=(LEFTMASK[xmf]&dst[xf])|(RIGHTMASK[xmf]&src[xf]);
             for (xi=xf+1; xi<xt; xi++) dst[xi]=src[xi];
             dst[xt]=(R2MASK[xmt]&dst[xt])|L2MASK[xmt]&src[xt];
            }
            dst+=PICO_TOTAL_X/2;
            src+=PICO_TOTAL_X/2;
        }
        for (; yi<=0; yi--) {
            if (xf==xt) {
             dst[xf]=(((LEFTMASK[xmf])|(R2MASK[xmt]))&dst[xf]);
            } else {
             unsigned char pattern;
             
             pattern=bgc*0x11;
             dst[xf]=(LEFTMASK[xmf]&dst[xf])|(RIGHTMASK[xmf]&pattern);
             for (xi=xf+1; xi<xt; xi++) dst[xi]=pattern;
             dst[xt]=(R2MASK[xmt]&dst[xt])|(L2MASK[xmt]&pattern);
            }
         dst+=PICO_TOTAL_X/2;
        }
    }
}

void picoDriverInit(void) {
    /* Clear Screen */
    picoFillRectRaw(0,0,PICO_X,PICO_Y,0);
}
