/*
 *  $Id: PTK_driver_8bpp.c,v 1.2 2000/12/03 12:14:28 tgkamp Exp $
 *
 *  picoTK 8 bit per pixel driver, 16 colors
 *
 *  Copyright (c) by Thomas Gallenkamp <tgkamp@users.sourceforge.net>
 *
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


#include <stdio.h>
#include "PTK.h"


/* Change the below to adapt driver - do this in the Makefile rather
   than here. The defaultsettings here are for compatibility with VGA
   hardware (640x480)
 */

#ifndef PICO_TOTAL_X
 #define PICO_TOTAL_X 640
#endif
#ifndef PICO_X
 #define PICO_X 640
#endif
#ifndef PICO_Y
 #define PICO_Y 480
#endif


int picoInfoSizeX(void)      { return PICO_X; }

int picoInfoSizeY(void)      { return PICO_Y; }

int picoInfoColorDepth(void) { return 8; }


inline void picoDrawPointRaw(int x, int y,int color) {
  unsigned char *p;
  fbuf[x+y*PICO_TOTAL_X]=color;
}

int picoDrawPixmapRaw (int x, int y, struct picoPixmap *pixmap,
                       unsigned char fgc, unsigned char bgc) {
    unsigned char *dst;
    unsigned char *src;
    int height;
    int width;
    int width_bytes;
    int xi, yi;
    int xm8,xm8m7;

    /*
     * We can only do depth 1 pixmaps
     */
    if (pixmap->type!=PICO_PIXMAP_1BPP) return 0;
   
    dst=fbuf+y*PICO_TOTAL_X+x;
    height=pixmap->height_msb*256+pixmap->height_lsb;
    width =pixmap->width_msb *256+pixmap->width_lsb;
    width_bytes=(width+7)/8;
    src=pixmap->data;

    for (yi=0; yi<height; yi++) {
     unsigned char *p;
     p=dst;
     for (xi=0; xi<width; ) {
      unsigned char data;
      int i;
      data=*(src++);
      for (i=0; (i<8) && xi<width; xi++,i++) {
        *(p++)=((data<<i) & 0x80) ? fgc:bgc;
      }/* i */
     } /* xi */
     dst+=PICO_TOTAL_X;
    } 
}

void picoFillRectRaw(int x, int y, int w, int h, int color) {
  unsigned char *dst;
  int xt;
  int xi;

  xt=x+w-1;
  for (h+=y; y<h; y++) {
   unsigned char mask;
   dst=fbuf+y*PICO_TOTAL_X+x;
   for (xi=x; xi<=xt; xi++) *(dst++)=color;
  }
}

void picoReverseRectRaw(int x, int y, int w, int h) {
    unsigned char *dst;
    int xt;
    int xi;

    xt=x+w-1;
    for (h+=y; y<h; y++) {
     unsigned char mask;
     dst=fbuf+y*PICO_TOTAL_X+x;
     for (xi=x; xi<=xt; xi++) *(dst++)^=255;
    }
}

int picoDrawTextRaw(int x, int y, char *txt, struct picoFont *font,
                    int fgc, int bgc) {
  int ch;
  int cont;
  int x_orig;
  
  x_orig=x;
  if (font==NULL) return 0;
  cont=1;
  while (ch=(unsigned char)*txt) {
    int height;
    int width;
    int width_in_bytes;
    int yi,xi,xp;
    unsigned char *src;
    unsigned char *dst;
    unsigned char lrmask;
    unsigned char xm2;
    unsigned char src_data;
    height=font->height;
    width=font->per_char[ch].width;
    if (x+width>PICO_X) {
        cont=0;
        break;
    }
    width_in_bytes=(width+7)/8;
    src =font->data+font->per_char[ch].offset_msb*256
                        +font->per_char[ch].offset_lsb;
    dst=fbuf+y*PICO_TOTAL_X+x;
      /* generic code */
    xm2=x&1;
    for (yi=0; yi<height; yi++) {
     unsigned char *p;
     p=dst;
     
     for (xi=0;xi<width; ) {
      unsigned char data;
      int i;
      data=*(src++);
      for (i=0; i<8 && xi<width; xi++, i++) {
         *(p++)=((data<<i) & 0x80) ? fgc:bgc;
      } /* rof i */
     } /* rof xi */
     dst+=PICO_TOTAL_X;
    } /* rof yi */
    x+=width;
    txt++;
  } /* elihw ch */
 return x-x_orig;
}


int picoScrollRaw(int x, int y, int w, int h, int delta, int bgc) {
    unsigned char *src;
    unsigned char *dst;
    int xi,yi;

    int xt;

    xt=x+w-1;
    
    if (delta>0) {
        dst=fbuf+y*PICO_TOTAL_X;
        src=dst+delta*PICO_TOTAL_X;
        for (yi=0; yi<h-delta; yi++) {
         for (xi=x; xi<=xt; xi++) dst[xi]=src[xi];
         dst+=PICO_TOTAL_X;
         src+=PICO_TOTAL_X;
        }
        for (; yi<h; yi++) {
         for (xi=x; xi<=xt; xi++) dst[xi]=bgc;
         dst+=PICO_TOTAL_X;
        }
    } else {
        /*
         * not yet tested ...
         */
        dst=fbuf+(y+h-1)*PICO_TOTAL_X;
        src=dst-delta*PICO_TOTAL_X;
        for (yi=h-1; yi>=h-delta; yi--) {
         for (xi=x; xi<=xt; xi++) dst[xi]=src[xi];
         dst+=PICO_TOTAL_X/2;
         src+=PICO_TOTAL_X/2;
        }
        for (; yi<=0; yi--) {
         for (xi=x; xi<=xt; xi++) dst[xi]=bgc;
         dst+=PICO_TOTAL_X/2;
        }
    }
}

void picoDriverInit(void) {
    /* Clear Screen */
    picoFillRectRaw(0,0,PICO_X,PICO_Y,0);
}
